import easy_tui as sTUI
sTUI.
from easy_tui import controls

hello = controls.Label(25, 3, "Hola, mundo")

nombre = controls.Entry(25, 4, "Nombre: ")
apellidos = controls.Entry(25, 5, "Apellidos: ")

def cambiaNombre():
    hello.text = f"Hola, {nombre.value} {apellidos.value}"

def limpiaTodo():
    nombre.value = ""
    apellidos.value = ""
    cambiaNombre()


f1 = controls.KeyButton(0, key=sTUI.F1, label="Reset", command=limpiaTodo)
f2 = controls.KeyButton(16,key=sTUI.F2, command=cambiaNombre, label="Saluda")


sTUI.loop()





from easy_tui import loop, LAST_COL, LAST_ROW
from calculator.views import *
from calculator.models import RomanCalculation
from simple_roman_numbers import RomanNumber

class Calculator:
    def __init__(self):
        f0 = (LAST_ROW - 20) // 2
        c0 = (LAST_COL - 17) // 2
        self.display = Display(fila=f0, col=c0)
        self.keyboard = Keyboard(output=self.receive, col=c0, fila = f0 + 4)
        self.calculation = RomanCalculation()
        self.resetDisplay = False
        self.keyboard_buffer = ""
        

    def receive(self, key):
        if key in tuple(self.calculation.operators):
            if self.keyboard_buffer:
                self.calculation.receive(self.keyboard_buffer)
                self.keyboard_buffer = ""
            self.calculation.receive(key)
            self.display.value = RomanNumber(self.calculation.display).lit
        else:
            buffer = self.keyboard_buffer + key
            try:
                value = RomanNumber(buffer)
                self.keyboard_buffer += key
                self.display.value = str(value)
            except ValueError:
                pass

    def start(self):
        loop()




from easy_tui import controls
import easy_tui as sTUI

teclas = ['M', '•', '/', 'C', 'D', '*', 'X', 'L', '-', 'I', 'V', '+']

class CalcButton(controls.KeyButton):
    def __init__(self, col, fila, key, command):
        super().__init__(col, fila, key, command)

    @controls.click_animate
    def draw(self):
        limit = "+-----+"
        middle = "|     |"
        sTUI.locate(self.col, self.fila, limit)
        sTUI.locate(self.col, self.fila+1, middle)
        sTUI.locate(self.col, self.fila+2, f"|  {self.label}  |")
        sTUI.locate(self.col, self.fila+3, middle)
        sTUI.locate(self.col, self.fila+4, limit)


class Display(controls.Label):
    def __init__(self, col, fila):
        super().__init__(col, fila, ">")
        self.value = ""

    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, v: str):
        self._value = v
        self.text = f"> {v:>15s}"

    def draw(self):
        sTUI.locate(self.col, self.fila,     "+-----------------+")
        sTUI.locate(self.col, self.fila + 1, "|                 |")
        sTUI.locate(self.col, self.fila + 2, f"|{self.text}|")
        sTUI.locate(self.col, self.fila + 3, "|                 |")
        sTUI.locate(self.col, self.fila + 4, "+-----------------+")
    

class Keyboard:
    def __init__(self, col=0, fila=0, output=None):
        teclas = ['M', '•', '/', 'C', 'D', '*', 'X', 'L', '-', 'I', 'V', '+']
        ix = 0
        f0 = fila
        c0 = col

        self.output = output

        def make_fn(key):
            return lambda: self.send(key)
            
        for fila in range(4):
            for col in range(3):
                button = CalcButton(c0 + col * 6, f0 + fila * 4, teclas[ix], make_fn(teclas[ix]))
                ix += 1

        

    def send(self, key):
        if self.output:
            self.output(key)
        






from simple_screen import pause, locate, STDSRC, A_UNDERLINE, A_REVERSE, Print, BACKSPACE, TAB, ENTER, STAB, key_map, key_labels
from abc import ABC, abstractmethod
from easy_tui import widgets, tabulatables, buttons, LAST_COL, LAST_ROW

key_list = list(key_map.values())

def click_animate(method):
    def wrapper(self, *args, **kwargs):
        if self.clicked:
            A_REVERSE.on()
            pause(10)
        result = method(self, *args, **kwargs)
        if self.clicked:
            A_REVERSE.off()
            self.clicked = False
        return result
    return wrapper

class Mixin(ABC):
    @abstractmethod
    def configure(self):
        pass

class Control(ABC):
    def __init__(self, col=0, fila=0, ancho=16, alto=1):
        self.col = col
        self.fila = fila
        self.ancho = ancho
        self.alto = alto

        widgets.append(self)
        if isinstance(self, Mixin):
            self.configure()

    @abstractmethod
    def draw(self, *args):
        pass

    """ @abstractmethod
    def update(self, *args):
        pass

    @abstractmethod
    def handle_input(self, key, *args):
        pass

    @abstractmethod
    def validate(self, value):
        pass """

class ButtonMixin(Mixin):
    def configure(self):
        buttons.append(self)

class TabulatableMixin(Mixin):
    _focused = False

    def configure(self):
        tabulatables.append(self)

    @property
    def focused(self):
        return self._focused
    
    def focus_in(self):
        self._focused = True

    def focus_out(self):
        self._focused = False

class Label(Control):
    def __init__(self, col=0, fila=0, text="Label"):
        super().__init__(col, fila)
        self.text = text
        
    def draw(self):
        locate(self.col, self.fila, self.text)


class Entry(Control, TabulatableMixin):
    def __init__(self, col=0, fila=0, label="", ancho=16):
        super().__init__(col, fila, ancho)

        self.label = label
        self.col_value = self.col + len(self.label) + 1
        self.buffer = []

    def draw(self):
        locate(self.col, self.fila, f"{self.label}")

        if self.focused:
            A_UNDERLINE.on()
        locate(self.col + len(self.label), self.fila, (self.value or ' '))
        A_UNDERLINE.off()
        Print(" ")

    @property
    def value(self):
        return ''.join(self.buffer)
    
    @value.setter
    def value(self, text: str):
        self.buffer = list(text)
    
    def handle_input(self, key):
        if key in BACKSPACE:
            if self.buffer:
                self.buffer.pop()
        elif key in (TAB, ENTER, STAB):
            return True
        elif key and key not in key_list:
            self.buffer.append(key)


class KeyButton(Control, ButtonMixin):
    def __init__(self, col=0, fila=LAST_ROW, key=None, command=None, label=""):
        super().__init__(col, fila)
        self.key = key
        strkey = key_labels.get(key, "!¿?")
        self.label = f"{strkey}: {label}" if label else strkey
        self._command = command or (lambda: True)
        self.clicked = False

    def click(self):
        self.clicked = True
        self.command()


    @property
    def command(self):
        return self._command
    
    @command.setter
    def command(self, f):
        self._command = f
    
    @click_animate
    def draw(self):
        locate(self.col, self.fila, f"| {self.label} |")






from collections import namedtuple
from simple_roman_numbers import RomanNumber

EQUAL_OPERATOR = "="

class Operation(namedtuple("operator",("symbol", "func"))):
    __slots__ = ()

    def resolve(self, v1, v2):
        return self.func(v1, v2)
    
Operators = [Operation("+", lambda a, b: a + b),
             Operation("-", lambda a, b: a - b),
             Operation("*", lambda a, b: a * b),
             Operation("/", lambda a, b: a / b),
             Operation("=", lambda a, b: None)]

class Calculation:
    operators = {op.symbol: op.func for op in Operators}

    def __init__(self, op1 = None, op2 = None, operator = None):
        self._op1 = op1
        self._op2 = op2
        self._operator = operator
        self._resolved = False
        self.display = 0

    def __resolve(self):
        operation = self.operators.get(self._operator)
        result = operation(self.op1, self.op2) if operation and self.op2 else None
        return result

    def add_number(self, value):
        if self._op1 is None:
            self._op1 = value
        elif self._op2 is None:
            self._op2 = value
        else:
            self._op1 = self._op2
            self._op2 = value

        self.display = self.op2 or self.op1
        self._resolved = False
    
    @property
    def op1(self):
        return self._op1
    

    @property
    def op2(self):
        return self._op2
    

    @property
    def operator(self):
        return self._operator
    
    @operator.setter
    def operator(self, value):
        if value not in list(self.operators):
            raise ValueError(f"{value} must be in ({','.join(self.operators)})")

        if self.op2 is None:
            self._operator = value
            return
        
        if value != EQUAL_OPERATOR:
            if not self._resolved:
                self.display = self.__resolve()
                self._op2 = self.display
            else:
                self._op1 = self._op2 = self.display
        
            self._operator = value
            self._resolved = True
        else:
            self.display = self.__resolve()
            if not self._resolved:
                self.add_number(self.display)
            else:
                self._op2 = self.display

            self._resolved = True
        
    def receive(self, value: str):
        if value in list(self.operators):
            self.operator = value
        else:
            self.add_number(float(value))

class RomanCalculation(Calculation):
    def receive(self, value: str):
        if value in list(self.operators):
            self.operator = value
        else:
            self.add_number(float(RomanNumber(value)))




from setuptools import setup, find_packages

setup(
    name="easy_tui",
    version="0.1.3",
    packages=find_packages(),
    include_package_data=True,
    description="A miniTUI with the ability to handle simple keyboard events in a terminal.\nIt's about avoiding all the noise when learning to program.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Mon Maldonado",
    author_email="monterdi@gmail.com",
    url="https://github.com/digestiveThinking/simple_screen",
    install_requires=[
        'simple_screen==0.1.14'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)



import simple_screen as ssc
from simple_screen.keys import *
LAST_ROW = ssc.DIMENSIONS.h - 1
LAST_COL = ssc.DIMENSIONS.w - 1


widgets = []
tabulatables = []
ix_tab = None
buttons = []

def change_tab(salto):
    global ix_tab
    tabulatables[ix_tab].focus_out()
    ix_tab = (ix_tab + salto) % len(tabulatables)
    tabulatables[ix_tab].focus_in()

def draw():
    ssc.cls()
    for control in widgets:
        control.draw()

@ssc.app
def loop():
    global ix_tab
    if tabulatables:
        tabulatables[0].focus_in()
        ix_tab = 0

    app_on = True
    while app_on:
        key = ssc.inkey()
        if key == ssc.ESC:
            app_on = False
        elif key in (ssc.TAB, ssc.ENTER):
            change_tab(1)
        elif key == ssc.STAB:
            change_tab(-1)

        for control in tabulatables:
            if control.focused:
                control.handle_input(key)

        for control in buttons:
            if control.key == key:
                control.click()

        draw()



    


from calculator.models import Calculation, RomanCalculation

def test_operation_simple():
    calculation = Calculation()
    assert calculation.display == 0

    calculation.receive('1')
    assert calculation.display == 1

    calculation.receive('+')
    assert calculation.display == 1

    calculation.receive('2')
    assert calculation.display == 2

def test_operations_repe_equals():
    calculation = Calculation()
    calculation.receive('1')
    calculation.receive('+')
    calculation.receive('2')

    calculation.receive('+')
    assert calculation.display == 3

    calculation.receive('+')
    assert calculation.display == 3

    calculation.receive('=')
    assert calculation.display == 6

    calculation.receive('=')
    assert calculation.display == 9
    
    
def test_operations_contiguous_operator():
    calculation = Calculation()
    calculation.receive('1')
    calculation.receive('+')
    calculation.receive('2')

    calculation.operator = '+'
    assert calculation.display == 3

    calculation.receive('5')
    assert calculation.display == 5

    calculation.operator = '-'
    assert calculation.display == 8

    calculation.receive('1')
    assert calculation.display == 1

    calculation.operator = '-'
    assert calculation.display == 7

    calculation.receive('3')
    assert calculation.display == 3

    calculation.operator = '='
    assert calculation.display == 4


def test_operation_equal_after_repe_operator():
    calculation = Calculation()
    calculation.receive('1')
    calculation.receive('+')
    calculation.receive('2')

    calculation.operator = '+'
    assert calculation.display == 3

    calculation.operator = '+'
    assert calculation.display == 3

    calculation.operator = '='
    assert calculation.display == 6

def test_operation_more_than_one():
    calculation = Calculation()
    calculation.receive('1')
    calculation.receive('+')
    calculation.receive('2')

    calculation.receive('+')
    assert calculation.display == 3

    calculation.receive('2')
    assert calculation.display == 2

    calculation.receive('-')
    assert calculation.display == 5

    calculation.receive('1')
    assert calculation.display == 1

    calculation.receive('*')
    assert calculation.display == 4

    calculation.receive('3')
    assert calculation.display == 3

    calculation.receive('=')
    assert calculation.display == 12

def test_roman_operation_simple():
    calculation = RomanCalculation()
    assert calculation.display == 0

    calculation.receive('I')
    assert calculation.display == 1

    calculation.receive('+')
    assert calculation.display == 1

    calculation.receive('II')
    assert calculation.display == 2





    


